--- ast_expr.y.orig	Sun Oct  5 22:55:26 2003
+++ ast_expr.y	Sun Oct  5 23:57:54 2003
@@ -30,8 +30,12 @@
 #define YYPARSE_PARAM kota
 #define YYLEX_PARAM kota
 
-/* #define ast_log fprintf
-#define LOG_WARNING stderr */
+#ifdef STANDALONE
+#define ast_log fprintf
+#define LOG_WARNING stderr
+#define LOG_DEBUG stderr
+#define LOG_ERROR stderr
+#endif
   
 enum valtype {
 	integer, numeric_string, string
@@ -61,7 +65,7 @@
 static int		is_zero_or_null __P((struct val *));
 static int		isstring __P((struct val *));
 static struct val	*make_integer __P((quad_t));
-static struct val	*make_str __P((const char *));
+static struct val	*make_str __P((const char *, const char *));
 static struct val	*op_and __P((struct val *, struct val *));
 static struct val	*op_colon __P((struct val *, struct val *));
 static struct val	*op_div __P((struct val *, struct val *));
@@ -145,19 +149,28 @@
 }
 
 static struct val *
-make_str (s)
-const char *s;
+make_str (q, p)
+const char *q;
+const char *p;
 {
 	struct val *vp;
 	size_t i;
 	int isint;
+	char *s;
 
+	if(p)
+		i = p - q;
+	else
+		i = strlen(q);
 	vp = (struct val *) malloc (sizeof (*vp));
-	if (vp == NULL || ((vp->u.s = strdup (s)) == NULL)) {
+	if(vp)
+		vp->u.s = malloc(i + 1);
+	if (vp == NULL || (vp->u.s == NULL)) {
 		ast_log(LOG_WARNING,"malloc() failed\n");
 		return(NULL);
 	}
-
+	strncpy(s = vp->u.s, q, i);
+	s[i] = '\0';
 	for(i = 1, isint = isdigit(s[0]) || s[0] == '-';
 	    isint && i < strlen(s);
 	    i++)
@@ -252,32 +265,76 @@
 ast_yylex (YYSTYPE *lvalp, struct parser_control *karoto)
 {
 	char *p;
+	int op;
+	char quote;
 
-	if (karoto->firsttoken==1) {
-		p=strtok_r(karoto->argv," ",&(karoto->ptrptr));
-		karoto->firsttoken=0;
-	} else {
-		p=strtok_r(NULL," ",&(karoto->ptrptr));
+	if(karoto->firsttoken == 1)
+	{
+		karoto->ptrptr = karoto->argv;
+		karoto->firsttoken = 0;
 	}
-
-	if (p==NULL) {
-		return (0);
+	op = -1;
+	/* Skip spaces */
+	for(p = karoto->ptrptr; *p == ' '; ++p);
+	karoto->ptrptr = p;
+	if(!*p)
+		return(0);
+	if(*p && (p[1] == '='))
+	{
+		switch(*p)
+		{
+			case '>':
+				op = GE;
+				p += 2;
+				break;
+			case '<':
+				op = LE;
+				p += 2;
+				break;
+			case '!':
+				op = NE;
+				p += 2;
+				break;
+		}
 	}
-
-
-	if (strlen (p) == 1) {
-		if (strchr ("|&=<>+-*/%:()", *p))
-			return (*p);
-	} else if (strlen (p) == 2 && p[1] == '=') {
-		switch (*p) {
-		case '>': return (GE);
-		case '<': return (LE);
-		case '!': return (NE);
+	else if(strchr("|&=<>+-*/%:()", *p))
+		op = *p++;
+	if(op < 0)
+	{
+		if((*p == '\'') || (*p == '"'))
+	        {
+			quote = *p;
+			for(++p; *p && (*p != quote); ++p)
+			{
+				if(*p == '\\')
+					++p;
+			}
+			if(*p)
+			{
+				++p;
+				op = TOKEN;
+			}
+			else
+				ast_log(LOG_WARNING, "Unbalanced %c: %s\n", quote, karoto->ptrptr);
+		}
+		else
+		{
+			for(; !strchr("|&=<>+-*/%:() ", *p); ++p)
+			{
+				if(*p == '\\')
+					++p;
+			}
+			op = TOKEN;
 		}
 	}
-
-	lvalp->val = make_str (p);
-	return (TOKEN);
+	if(op >= 0)
+		lvalp->val = make_str(karoto->ptrptr, p);
+#if YYDEBUG
+//	if(yydebug)
+		ast_log(LOG_DEBUG, "yylex returns %.*s as %d\n", p - karoto->ptrptr, karoto->ptrptr, op);
+#endif
+	karoto->ptrptr = p;
+	return(op);
 }
 
 static int
@@ -328,6 +385,9 @@
 
 int main(int argc,char **argv) {
 	char *s;
+#if YYDEBUG
+	yydebug = 1;
+#endif
 
 	s=ast_expr(argv[1]);
 
@@ -714,7 +774,7 @@
 	if (regexec(&rp, a->u.s, (size_t)2, rm, 0) == 0 && rm[0].rm_so == 0) {
 		if (rm[1].rm_so >= 0) {
 			*(a->u.s + rm[1].rm_eo) = '\0';
-			v = make_str (a->u.s + rm[1].rm_so);
+			v = make_str (a->u.s + rm[1].rm_so, (char *)NULL);
 
 		} else {
 			v = make_integer ((quad_t)(rm[0].rm_eo - rm[0].rm_so));
@@ -723,7 +783,7 @@
 		if (rp.re_nsub == 0) {
 			v = make_integer ((quad_t)0);
 		} else {
-			v = make_str ("");
+			v = make_str ("", (char *)NULL);
 		}
 	}
 
